"""Detect from trajectory."""

import pickle
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from enum import Enum, auto
from typing import List, Optional, Tuple, Union, cast

import numpy as np
from ase import Atom, Atoms
from openbabel import openbabel

from mddatasetbuilder.dps import dps as connectmolecule


class Detect(metaclass=ABCMeta):
    """Detect structures from file(s)."""

    def __init__(self, filename, atomname, pbc, errorlimit=None, errorfilename=None):
        self.filename = filename
        self.atomname = atomname
        self.pbc = pbc
        self.errorlimit = errorlimit
        self.errorfilename = errorfilename
        self.steplinenum = self._readN()

    @abstractmethod
    def _readN(self) -> int:
        pass

    @abstractmethod
    def readatombondtype(self, item) -> Tuple[dict, int]:
        """Read bond types of atoms such as C1111."""
        pass

    @abstractmethod
    def readmolecule(self, lines) -> Tuple[List[List[int]], Optional[Atoms]]:
        """Read molecules."""
        pass

    @staticmethod
    def gettype(inputtype):
        """Get the class for the input file type."""
        if inputtype == "bond":
            detectclass = DetectBond
        elif inputtype == "dump":
            detectclass = DetectDump
        else:
            raise RuntimeError("Wrong input file type")
        return detectclass


class DetectBond(Detect):
    """Detect from the LAMMPS bond file."""

    def _readN(self):
        """Read bondfile N, which should be at very beginning."""
        N = None
        atomtype = None
        stepaindex = None
        stepbindex = None
        # copy from reacnetgenerator on 2018-12-15
        with open(
            self.filename if isinstance(self.filename, str) else self.filename[0]
        ) as f:
            iscompleted = False
            for index, line in enumerate(f):
                if line.startswith("#"):
                    if line.startswith("# Number of particles"):
                        if iscompleted:
                            stepbindex = index
                            break
                        else:
                            iscompleted = True
                            stepaindex = index
                        N = [int(s) for s in line.split() if s.isdigit()][0]
                        atomtype = np.zeros(N, dtype=int)
                else:
                    s = line.split()
                    assert atomtype is not None
                    atomtype[int(s[0]) - 1] = int(s[1])
        if stepaindex is None or stepbindex is None or N is None or atomtype is None:
            raise RuntimeError("The bond file is not completed")
        steplinenum = stepbindex - stepaindex
        self._N = N
        self.atomtype = atomtype
        self.atomnames = self.atomname[self.atomtype - 1]
        return steplinenum

    def readatombondtype(self, item):
        """Read bond orders of atoms.

        Parameters
        ----------
        item : tuple
            (step, lines), _

        Returns
        -------
        dict
            dict of bond orders
        int
            the step index
        """
        # copy from reacnetgenerator on 2018-12-15
        (step, lines), _ = item
        d = defaultdict(list)
        for line in lines:
            if line:
                if line[0] != "#":
                    s = line.split()
                    atombond = sorted(
                        max(1, round(float(x)))
                        for x in s[4 + int(s[2]) : 4 + 2 * int(s[2])]
                    )
                    d[pickle.dumps((self.atomnames[int(s[0]) - 1], atombond))].append(
                        int(s[0])
                    )
        return d, step

    def readmolecule(self, lines) -> Tuple[List[List[int]], Optional[Atoms]]:
        """Return molecules from lines.

        Parameters
        ----------
        lines : list of strs
            Lines of LAMMPS bond files.

        Returns
        -------
        molecules: list
            Indexes of atoms in molecules.
        None
            None
        """
        # copy from reacnetgenerator on 2018-12-15
        bond: List[Optional[List[int]]] = [None] * self._N
        for line in lines:
            if line:
                if not line.startswith("#"):
                    s = line.split()
                    bond[int(s[0]) - 1] = [int(x) - 1 for x in s[3 : 3 + int(s[2])]]
        bond_ = cast(List[List[int]], bond)
        molecules = connectmolecule(bond_)
        return molecules, None


class DetectDump(Detect):
    """Detect from the dump file."""

    def _readN(self):
        # copy from reacnetgenerator on 2018-12-15
        iscompleted = False
        N = None
        atomtype = None
        stepaindex = None
        stepbindex = None
        with open(
            self.filename if isinstance(self.filename, str) else self.filename[0]
        ) as f:
            linecontent = None
            for index, line in enumerate(f):
                if line.startswith("ITEM:"):
                    linecontent = self.LineType.linecontent(line)
                    if linecontent == self.LineType.ATOMS:
                        keys = line.split()
                        self.id_idx = keys.index("id") - 2
                        self.tidx = keys.index("type") - 2
                        self.xidx = keys.index("x") - 2
                        self.yidx = keys.index("y") - 2
                        self.zidx = keys.index("z") - 2
                else:
                    if linecontent is None:
                        raise RuntimeError("No ITEM: in the dump file")
                    elif linecontent == self.LineType.NUMBER:
                        if iscompleted:
                            stepbindex = index
                            break
                        else:
                            iscompleted = True
                            stepaindex = index
                        N = int(line.split()[0])
                        atomtype = np.zeros(N, dtype=int)
                    elif linecontent == self.LineType.ATOMS:
                        s = line.split()
                        assert atomtype is not None
                        atomtype[int(s[self.id_idx]) - 1] = int(s[self.tidx])
        if stepaindex is None or stepbindex is None or N is None or atomtype is None:
            raise RuntimeError("The dump file is not completed")
        steplinenum = stepbindex - stepaindex
        self._N = N
        self.atomtype = atomtype
        self.atomnames = self.atomname[self.atomtype - 1]
        return steplinenum

    def readatombondtype(self, item):
        """Read bond orders of atoms.

        Parameters
        ----------
        item : tuple
            (step, lines), _

        Returns
        -------
        dict
            dict of bond orders
        int
            the step index
        """
        (step, lines), needlerror = item
        lerror: Optional[Union[np.ndarray, List[float]]] = None
        d = defaultdict(list)
        step_atoms, ids = self.readcrd(lines)
        if needlerror:
            trajline, errorline = lines
            lerror = np.fromstring(errorline, dtype=float, sep=" ")[7:]
            lerror = [x for (y, x) in sorted(zip(ids, lerror))]
        level = self._crd2bond(step_atoms, readlevel=True)
        for i, (n, l) in enumerate(zip(self.atomnames, level)):
            if lerror is None or (
                self.errorlimit is not None and lerror[i] > self.errorlimit
            ):
                # Note that atom id starts from 1
                d[pickle.dumps((n, sorted(l)))].append(i + 1)
        return d, step

    def readmolecule(self, lines) -> Tuple[List[List[int]], Optional[Atoms]]:
        """Return molecules from lines.

        Parameters
        ----------
        lines : list of strs
            Lines of LAMMPS bond files.

        Returns
        -------
        molecules: list
            Indexes of atoms in molecules.
        step_atoms: ase.Atoms
            The atoms of the frame.
        """
        step_atoms, _ = self.readcrd(lines)
        bond = self._crd2bond(step_atoms, readlevel=False)
        molecules = connectmolecule(bond)
        # return atoms as well
        return molecules, step_atoms

    @classmethod
    def _crd2bond(cls, step_atoms, readlevel):
        # copy from reacnetgenerator on 2019/4/13
        # updated on 2019/10/11
        atomnumber = len(step_atoms)
        # Use openbabel to connect atoms
        mol = openbabel.OBMol()
        mol.BeginModify()
        for idx, (num, position) in enumerate(
            zip(step_atoms.get_atomic_numbers(), step_atoms.positions)
        ):
            a = mol.NewAtom(idx)
            a.SetAtomicNum(int(num))
            a.SetVector(*position)
        # Apply period boundry conditions
        # openbabel#1853, supported in v3.1.0
        if step_atoms.pbc.any():
            cell = step_atoms.cell
            uc = openbabel.OBUnitCell()
            uc.SetData(
                openbabel.vector3(cell[0][0], cell[0][1], cell[0][2]),
                openbabel.vector3(cell[1][0], cell[1][1], cell[1][2]),
                openbabel.vector3(cell[2][0], cell[2][1], cell[2][2]),
            )
            mol.CloneData(uc)
            mol.SetPeriodicMol()
        mol.ConnectTheDots()
        # when readlevel is False, bond is used to store connected atoms
        # otherwise, bondlevel is used to store bond orders
        bond = [[] for i in range(atomnumber)]
        if readlevel:
            mol.PerceiveBondOrders()
        mol.EndModify()
        for b in openbabel.OBMolBondIter(mol):
            s1 = b.GetBeginAtom().GetId()
            s2 = b.GetEndAtom().GetId()
            if not readlevel:
                bond[s1].append(s2)
                bond[s2].append(s1)
            else:
                level = b.GetBondOrder()
                bond[s1].append(level)
                bond[s2].append(level)
        return bond

    def readcrd(self, item) -> Tuple[Atoms, List[int]]:
        """Only this function can read coordinates."""
        lines = item
        # box information
        ss = []
        step_atoms = []
        ids = []
        linecontent = None
        for line in lines:
            if line:
                if line.startswith("ITEM:"):
                    linecontent = self.LineType.linecontent(line)
                else:
                    if linecontent is None:
                        raise RuntimeError("No ITEM: in the dump file")
                    elif linecontent == self.LineType.ATOMS:
                        s = line.split()
                        ids.append(int(s[self.id_idx]))
                        step_atoms.append(
                            Atom(
                                self.atomname[int(s[self.tidx]) - 1],
                                (
                                    float(s[self.xidx]),
                                    float(s[self.yidx]),
                                    float(s[self.zidx]),
                                ),
                            )
                        )
                    elif linecontent == self.LineType.BOX:
                        s = line.split()
                        ss.append(list(map(float, s)))
        # box information to 3x3 cell
        ss = np.array(ss)
        if ss.shape[1] > 2:
            xy = ss[0][2]
            xz = ss[1][2]
            yz = ss[2][2]
        else:
            xy, xz, yz = 0.0, 0.0, 0.0
        xlo = ss[0][0] - min(0.0, xy, xz, xy + xz)
        xhi = ss[0][1] - max(0.0, xy, xz, xy + xz)
        ylo = ss[1][0] - min(0.0, yz)
        yhi = ss[1][1] - max(0.0, yz)
        zlo = ss[2][0]
        zhi = ss[2][1]
        boxsize = np.array(
            [[xhi - xlo, 0.0, 0.0], [xy, yhi - ylo, 0.0], [xz, yz, zhi - zlo]]
        )
        # sort by ID
        step_atoms_ = [x for (y, x) in sorted(zip(ids, step_atoms))]
        step_atoms = Atoms(step_atoms_, cell=boxsize, pbc=self.pbc)
        return step_atoms, ids

    class LineType(Enum):
        """Line type in the LAMMPS dump files."""

        TIMESTEP = auto()
        ATOMS = auto()
        NUMBER = auto()
        BOX = auto()
        OTHER = auto()

        @classmethod
        def linecontent(cls, line):
            """Return line content."""
            if line.startswith("ITEM: TIMESTEP"):
                return cls.TIMESTEP
            if line.startswith("ITEM: ATOMS"):
                return cls.ATOMS
            if line.startswith("ITEM: NUMBER OF ATOMS"):
                return cls.NUMBER
            if line.startswith("ITEM: BOX"):
                return cls.BOX
            return cls.OTHER



"""Test."""

import hashlib
import json
import math
import os
import tempfile
from pathlib import Path

import pytest
import requests
from tqdm.auto import tqdm

import mddatasetbuilder
import mddatasetbuilder.deepmd
import mddatasetbuilder.qmcalc
from mddatasetbuilder._logger import logger

this_directory = os.getcwd()
with open(Path(__file__).parent / "test.json") as f:
    test_params = json.load(f)


class TestMDDatasetBuilder:
    """Test MDDatasetBuilder."""

    @pytest.fixture(params=test_params)
    def datasetbuilder(self, request):
        """DatasetBuilder fixture."""
        folder = tempfile.mkdtemp(prefix="testfiles-", dir=this_directory)
        logger.info(f"Folder: {folder}:")
        os.chdir(folder)
        testparms = request.param
        # download bonds.reaxc and dump.reaxc
        for fileparms in (
            (testparms["bondfile"], testparms["dumpfile"])
            if "bondfile" in testparms
            else (testparms["dumpfile"],)
        ):
            self._download_file(
                fileparms["url"], fileparms["filename"], fileparms["sha256"]
            )

        return mddatasetbuilder.DatasetBuilder(
            bondfilename=testparms["bondfile"]["filename"]
            if "bondfile" in testparms
            else None,
            dumpfilename=testparms["dumpfile"]["filename"],
            atomname=testparms["atomname"],
            n_clusters=testparms["size"],
            dataset_name=testparms["dataset_name"],
            stepinterval=testparms["stepinterval"]
            if "stepinterval" in testparms
            else 1,
        )

    def test_datasetbuilder(self, datasetbuilder):
        """Test DatasetBuilder."""
        datasetbuilder.builddataset()
        assert os.path.exists(datasetbuilder.gjfdir)
        for ii in os.listdir(datasetbuilder.gjfdir):
            mddatasetbuilder.qmcalc.qmcalc(os.path.join(datasetbuilder.gjfdir, ii))
        mddatasetbuilder.deepmd.PrepareDeePMD(datasetbuilder.gjfdir).preparedeepmd()

    def _download_file(self, urls, pathfilename, sha256):
        times = 0
        # download if not exists
        while times < 3:
            if os.path.isfile(pathfilename) and self._checksha256(pathfilename, sha256):
                break
            Path(pathfilename).resolve().parent.mkdir(parents=True, exist_ok=True)

            # from https://stackoverflow.com/questions/16694907
            if not isinstance(urls, list):
                urls = [urls]
            for url in urls:
                try:
                    logger.info(f"Try to download {pathfilename} from {url}")
                    r = requests.get(url, stream=True)
                    break
                except requests.exceptions.RequestException as e:
                    logger.warning(e)
                    logger.warning("Request Error.")
            else:
                logger.error(f"Cannot download {pathfilename}.")
                raise OSError(f"Cannot download {pathfilename}.")

            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024
            with open(pathfilename, "wb") as f:
                for chunk in tqdm(
                    r.iter_content(chunk_size=1024),
                    total=math.ceil(total_size // block_size),
                    unit="KB",
                    unit_scale=True,
                    desc=f"Downloading {pathfilename}...",
                    disable=None,
                ):
                    if chunk:
                        f.write(chunk)
        else:
            logger.error("Retry too much times.")
            raise OSError("Retry too much times.")
        return pathfilename

    @staticmethod
    def _checksha256(filename, sha256_check):
        if not os.path.isfile(filename):
            return
        h = hashlib.sha256()
        b = bytearray(128 * 1024)
        mv = memoryview(b)
        with open(filename, "rb", buffering=0) as f:
            for n in iter(lambda: f.readinto(mv), 0):
                h.update(mv[:n])
        sha256 = h.hexdigest()
        logger.info(f"SHA256 of {filename}: {sha256}")
        if sha256 == sha256_check:
            return True
        logger.warning("SHA256 is not correct.")
        return False



"""Utils."""

import itertools
import pickle
from multiprocessing import Pool, Semaphore
from typing import BinaryIO, List, TypeVar, Union, overload

import lz4.frame
from tqdm.auto import tqdm

from ._logger import logger


def multiopen(
    pool,
    func,
    l,
    semaphore=None,
    nlines=None,
    unordered=True,
    return_num=False,
    start=0,
    extra=None,
    interval=None,
    bar=True,
    desc=None,
    unit="it",
    total=None,
):
    """Return an interated object for process a file with multiple processors.

    Parameters
    ----------
    pool : multiprocessing.Pool
        The pool for multiprocessing.
    func : function
        The function to process lines.
    l : File object
        The file object.
    semaphore : multiprocessing.Semaphore, optional, default: None
        The semaphore to acquire. If None (default), the object will be passed
        without control.
    nlines : int, optional, default: None
        The number of lines to pass to the function each time. If None (default),
        only one line will be passed to the function.
    unordered : bool, optional, default: True
        Whether the process can be unordered.
    return_num : bool, optional, default: False
        If True, adds a counter to an iterable.
    start : int, optional, default: 0
        The start number of the counter.
    extra : object, optional, default: None
        The extra object passed to the item.
    interval : int, optional, default: None
        The interval of items that will be passed to the function. For example,
        if set to 10, a item will be passed once every 10 items and others will
        be dropped.
    bar : bool, optional, default: True
        If True, show a tqdm bar for the iteration.
    desc : str, optional, default: None
        The description of the iteration shown in the bar.
    unit : str, optional, default: it
        The unit of the iteration shown in the bar.
    total : int, optional, default: None
        The total number of the iteration shown in the bar.

    Returns
    -------
    object
        An object that can be iterated.
    """
    obj = l
    if nlines:
        obj = itertools.zip_longest(*[obj] * nlines)
    if interval:
        obj = itertools.islice(obj, 0, None, interval)
    if return_num:
        obj = enumerate(obj, start)
    if semaphore:
        obj = produce(semaphore, obj, extra)
    if unordered:
        obj = pool.imap_unordered(func, obj, 100)
    else:
        obj = pool.imap(func, obj, 100)
    if bar:
        obj = tqdm(obj, desc=desc, unit=unit, total=total, disable=None)
    return obj


def produce(semaphore, plist, parameter):
    """Prevent large memory usage due to slow IO."""
    for item in plist:
        semaphore.acquire()
        if parameter is not None:
            item = (item, parameter)
        yield item


def compress(x: Union[str, bytes]) -> bytes:
    """Compress the line.

    This function reduces IO overhead to speed up the program. The functions will
    use lz4 to compress, since lz4 has better performance that any others.
    The compressed format is size + data + size + data + ..., where size is a 64-bit
    little-endian integer.

    Parameters
    ----------
    x : str or bytes
        The line to compress.

    Returns
    -------
    bytes
        The compressed line, with a linebreak in the end.
    """
    if isinstance(x, str):
        x = x.encode()
    compress_block = lz4.frame.compress(x, compression_level=0)
    length_bytes = len(compress_block).to_bytes(64, byteorder="little")
    return length_bytes + compress_block


def decompress(x: bytes, isbytes: bool = False) -> Union[str, bytes]:
    """Decompress the line.

    Parameters
    ----------
    x : bytes
        The line to decompress.
    isbytes : bool, optional, default: False
        If the decompressed content is bytes. If not, the line will be decoded.

    Returns
    -------
    str or bytes
        The decompressed line.
    """
    x = lz4.frame.decompress(x[64:])
    if isbytes:
        return x
    return x.decode()


def read_compressed_block(f: BinaryIO):
    """Read compressed binary file, assuming the format is size + data + size + data + ...

    Parameters
    ----------
    f : fileObject
        The file object to read.

    Yields
    ------
    data: bytes
        The compressed block.
    """
    while True:
        sizeb = f.read(64)
        if not sizeb:
            break
        size = int.from_bytes(sizeb, byteorder="little")
        yield sizeb + f.read(size)


def listtobytes(x):
    """Convert an object to a compressed line.

    Parameters
    ----------
    x : object
        The object to convert, such as numpy.ndarray.

    Returns
    -------
    bytes
        The compressed line.
    """
    return compress(pickle.dumps(x))


def bytestolist(x):
    """Convert a compressed line to an object.

    Parameters
    ----------
    x : bytes
        The compressed line.

    Returns
    -------
    object
        The decompressed object.
    """
    return pickle.loads(decompress(x, isbytes=True))  # type: ignore


def run_mp(nproc, **arg):
    """Process a file with multiple processors.

    Parameters
    ----------
    nproc : int
        The number of processors to be used.
    **kwargs : dict, optional
        Other parameters can be found in the `multiopen` method.

    Yields
    ------
    object
        The yielded object from the `multiopen` method.

    See Also
    --------
    multiopen
    """
    pool = Pool(nproc, maxtasksperchild=1000)
    semaphore = Semaphore(nproc * 150)
    try:
        results = multiopen(pool=pool, semaphore=semaphore, **arg)
        for item in results:
            yield item
            semaphore.release()
    except:
        logger.exception("run_mp failed")
        pool.terminate()
        raise
    else:
        pool.close()
    finally:
        pool.join()


T = TypeVar("T")


@overload
def must_be_list(obj: List[T]) -> List[T]: ...
@overload
def must_be_list(obj: T) -> List[T]: ...


def must_be_list(obj: Union[T, List[T]]) -> List[T]:
    """Convert a object to a list if the object is not a list.

    Parameters
    ----------
    obj : Object
        The object to convert.

    Returns
    -------
    obj: list
        If the input object is not a list, returns a list that only contains that
        object. Otherwise, returns that object.
    """
    if isinstance(obj, list):
        return obj
    return [obj]



"""Test detecting bonds."""

import numpy as np
from ase import Atoms

from mddatasetbuilder.detect import DetectDump


def test_bond_pbc():
    """Test detecting bonds under PBC condition."""
    atoms = Atoms(
        "O2",
        positions=[[0.0, 0.0, 0.0], [19.0, 19.0, 19.0]],
        pbc=True,
        cell=np.diag([20.0, 20.0, 20.0]),
    )
    bonds = DetectDump._crd2bond(atoms, False)
    assert bonds == [[1], [0]]
    levels = DetectDump._crd2bond(atoms, True)
    assert levels == [[1], [1]]


def test_bond_nopbc():
    """Test detecting bonds under non-PBC condition."""
    atoms = Atoms(
        "O2",
        positions=[[0.0, 0.0, 0.0], [19.0, 19.0, 19.0]],
        pbc=False,
        cell=np.diag([20.0, 20.0, 20.0]),
    )
    bonds = DetectDump._crd2bond(atoms, False)
    assert bonds == [[], []]
    levels = DetectDump._crd2bond(atoms, True)
    assert levels == [[], []]



"""QM calculation with Gaussian."""

import argparse
import os
from typing import Optional

from gaussianrunner import GaussianRunner


def qmcalc(gjfdir, command="g16", cpu_num: Optional[int] = None):
    """QM Calculation."""
    gjflist = [
        os.path.join(gjfdir, filename)
        for filename in os.listdir(gjfdir)
        if filename.endswith(".gjf")
    ]
    GaussianRunner(command=command, cpu_num=cpu_num).runGaussianInParallel(
        "GJF", gjflist
    )


def _commandline():
    parser = argparse.ArgumentParser(description="QM Calculator")
    parser.add_argument("-d", "--dir", help="Dataset dirs", required=True)
    parser.add_argument(
        "-c", "--command", help="Gaussian command, default is g16", default="g16"
    )
    parser.add_argument(
        "-n",
        "--ncpus",
        help="Number of CPU cores used, if not set, all CPU cores are used",
        default=None,
        type=int,
    )
    args = parser.parse_args()
    qmcalc(gjfdir=args.dir, command=args.command)



"""Gaussian logs to DeePMD data files."""

import argparse
import json
import os
import random
from multiprocessing import Pool

import dpdata
import numpy as np
from tqdm.auto import tqdm


class PrepareDeePMD:
    """Prepare DeePMD training files."""

    def __init__(
        self,
        data_path,
        deepmd_dir="data",
        jsonfilenumber=1,
        fmt="gaussian/log",
        suffix=".log",
    ):
        """Init the class."""
        self.data_path = data_path
        self.deepmd_dir = deepmd_dir
        self.system_paths = []
        self.batch_size = []
        self.fmt = fmt
        self.suffix = suffix
        self.jsonfilenames = [
            os.path.join(f"train{i}", f"train{i}.json") for i in range(jsonfilenumber)
        ]

    def preparedeepmd(self):
        """Prepare the dataset."""
        self._searchpath()
        for jsonfilename in self.jsonfilenames:
            self._writejson(jsonfilename)

    def _searchpath(self):
        logfiles = []
        for root, _, files in tqdm(
            os.walk(self.data_path, followlinks=True), disable=None
        ):
            for logfile in files:
                if logfile.endswith(self.suffix):
                    logfiles.append(os.path.join(root, logfile))
        multi_systems = dpdata.MultiSystems()
        with Pool() as pool:
            for system in pool.imap_unordered(
                self._preparedeepmdforLOG, tqdm(logfiles, disable=None)
            ):
                multi_systems.append(system)
        multi_systems.to_deepmd_npy(self.deepmd_dir)  # type: ignore
        for formula, system in multi_systems.systems.items():
            self.system_paths.append(os.path.join(self.deepmd_dir, formula))
            self.batch_size.append(
                min(
                    max(32 // (system["coords"].shape[1]), 1), system["coords"].shape[0]
                )
            )
        self.atomname = multi_systems.atom_names

    def _preparedeepmdforLOG(self, logfilename):
        system = dpdata.LabeledSystem(logfilename, fmt=self.fmt)
        atom_pref_file = os.path.splitext(logfilename)[0] + ".atom_pref.npy"
        if os.path.exists(atom_pref_file):
            system.data["atom_pref"] = np.load(atom_pref_file)
        return system

    def _writejson(self, jsonfilename):
        jsonpath = os.path.dirname(jsonfilename)
        deepmd_json = {
            "model": {
                "type_map": self.atomname,
                "descriptor": {
                    "type": "se_a",
                    "sel": "auto",
                    "rcut_smth": 1.0,
                    "rcut": 6.0,
                    "neuron": [25, 50, 100],
                    "resnet_dt": False,
                    "axis_neuron": 12,
                    "seed": random.randint(0, 2**32),
                },
                "fitting_net": {
                    "neuron": [240, 240, 240],
                    "resnet_dt": True,
                    "seed": random.randint(0, 2**32),
                },
            },
            "learning_rate": {
                "type": "exp",
                "start_lr": 0.0005,
                "decay_steps": 20000,
                "decay_rate": 0.96,
            },
            "loss": {
                "start_pref_e": 0.2,
                "limit_pref_e": 0.2,
                "start_pref_f": 1000,
                "limit_pref_f": 1,
                "start_pref_v": 0,
                "limit_pref_v": 0,
            },
            "training": {
                "systems": [
                    os.path.relpath(path, jsonpath) for path in self.system_paths
                ],
                "set_prefix": "set",
                "stop_batch": 4000000,
                "batch_size": "auto",
                "seed": random.randint(0, 2**32),
                "disp_file": "lcurve.out",
                "disp_freq": 1000,
                "numb_test": 1,
                "save_freq": 1000,
                "save_ckpt": "./model.ckpt",
                "disp_training": True,
                "time_training": True,
            },
        }
        os.makedirs(jsonpath, exist_ok=True)
        with open(jsonfilename, "w") as f:
            json.dump(deepmd_json, f)


def _commandline():
    parser = argparse.ArgumentParser(description="Prepare DeePMD data")
    parser.add_argument(
        "-d", "--dir", help="Dataset dirs, default is data", default="data"
    )
    parser.add_argument(
        "-p",
        "--path",
        help="Gaussian LOG file path, e.g. dataset_md_GJF",
        required=True,
    )
    parser.add_argument(
        "-n", "--number", type=int, default=1, help="The number of train.json"
    )
    args = parser.parse_args()
    PrepareDeePMD(
        data_path=args.path, deepmd_dir=args.dir, jsonfilenumber=args.number
    ).preparedeepmd()



"""MDDatasetBuilder.

Run 'datasetbuilder -h' for more details.

Please cite
-----------
Complex reaction processes in combustion unraveled by neural network-based
molecular dynamics simulation, Nature Communications, 11, 5713 (2020).
"""

__author__ = "Jinzhe Zeng"
__email__ = "jzzeng@stu.ecnu.edu.cn"
__update__ = "2019-10-17"
__date__ = "2018-07-18"

import argparse
import gc
import itertools
import os
import pickle
import tempfile
import time
from collections import Counter, defaultdict
from typing import List, Optional

import numpy as np
from ase.atoms import Atoms
from ase.data import atomic_numbers
from ase.io import write as write_xyz
from sklearn import preprocessing
from sklearn.cluster import MiniBatchKMeans

from ._logger import logger
from ._version import version as __version__
from .detect import Detect, DetectDump
from .utils import (
    bytestolist,
    listtobytes,
    must_be_list,
    read_compressed_block,
    run_mp,
)


class DatasetBuilder:
    r"""Dataset Builder.

    Parameters
    ----------
    atomname: list, optional, default=['C', 'H', 'O']
        Atom names.
    clusteratom: list, optional, default=None
        Cluster elements. If None (default), all elements will be clustered.
    bondfilename: str, optional, default=None
        The filename of LAMMPS bond file. If None (default), bond files will
        not be used.
    dumpfilename: str, optional, default="dump.reaxc"
        The filename of LAMMPS dump file.
    dataset_name: str, optional, default="md"
        The name of dataset, which will be part of filenames of dataset.
    cutoff: float, optional, default=5.
        The cutoff for taking clusters.
    stepinterval: int, optional, default=1
        The interval for taking frames.
    n_clusters: int, optional, default=10000
        The maximum number of clusters for each atom type.
    n_each: int, optional, default=1
        The numbers of structures taken from each cluster.
    qmkeywords: str, optional, default="%nproc=4\n#force mn15/6-31g(d,p)"
        Gaussian keywords.
    nproc: int, optional, default=None
        The number of processors for building the dataset. If None, all processors
        will be used.
    pbc: bool, optional, default=True
        If True (default), apply the periodic boundary conditions (PBC).
    fragment: bool, optional, default=False
        Use `Guess=Fragment` for Gaussian calculation. See https://gaussian.com/guess/?tabid=1#Guess_keyword__Fragment_option
        for details.
    errorfilename: str, optional, default=None
        The atomic model deviation file of DeePMD-kit. If None, no file will be used.
    errorlimit: float, optional, default=0.
        The lower bound of the model deviation. The atom will be considered "accurate"
        if the atomic model deviation is less than this value.
    atom_pref: bool, optional, default=False
        (Deprecated) Generator atom_pref information for each cluster.
    """

    def __init__(
        self,
        atomname=None,
        clusteratom=None,
        bondfilename=None,
        dumpfilename="dump.reaxc",
        dataset_name="md",
        cutoff=5.0,
        stepinterval=1,
        n_clusters=10000,
        n_each=1,
        qmkeywords="%nproc=4\n#force mn15/6-31g(d,p) Geom=PrintInputOrient",
        nproc=None,
        pbc=True,
        fragment=False,
        errorfilename: Optional[List[str]] = None,
        errorlimit=0.0,
        atom_pref=False,
    ):
        """Init the builder."""
        print(__doc__)
        print(f"Author:{__author__}  Email:{__email__}")
        atomname = np.array(atomname) if atomname else np.array(["C", "H", "O"])
        self.crddetector = Detect.gettype("dump")(
            filename=dumpfilename,
            atomname=atomname,
            pbc=pbc,
            errorfilename=errorfilename,
            errorlimit=errorlimit,
        )
        if bondfilename is None:
            self.bonddetector = self.crddetector
        else:
            self.bonddetector = Detect.gettype("bond")(
                filename=bondfilename, atomname=atomname, pbc=pbc
            )

        self.dataset_dir = f"dataset_{dataset_name}"
        self.xyzfilename = dataset_name
        self.clusteratom = clusteratom if clusteratom else atomname
        self.atombondtype = []
        self.stepinterval = stepinterval
        if nproc:
            self.nproc = nproc
        else:
            try:
                self.nproc = len(os.sched_getaffinity(0))
            except AttributeError:
                # macos and windows
                self.nproc = os.cpu_count()
        self.cutoff = cutoff
        self.n_clusters = n_clusters
        self.n_each = n_each
        self.writegjf = True
        self.gjfdir = f"{self.dataset_dir}_gjf"
        self.qmkeywords = must_be_list(qmkeywords)
        self.fragment = fragment
        self._coulumbdiag = {
            symbol: atomic_numbers[symbol] ** 2.4 / 2 for symbol in atomname
        }
        self._nstructure = 0
        self.bondtyperestore = {}
        self.errorfilename = errorfilename
        self.atom_pref = atom_pref

    def builddataset(self, writegjf=True):
        """Build a dataset.

        Parameters
        ----------
        writegjf : bool, optional, default=True
            Write gjf files.
        """
        self.writegjf = writegjf
        timearray = [time.time()]
        with tempfile.TemporaryDirectory() as self.trajatom_dir:
            for runstep in range(3):
                if runstep == 0:
                    self._readtimestepsbond()
                elif runstep == 1:
                    with open(
                        os.path.join(self.trajatom_dir, "chooseatoms"), "wb"
                    ) as f:
                        for bondtype in self.atombondtype:
                            self._writecoulumbmatrix(bondtype, f)
                            gc.collect()
                elif runstep == 2:
                    os.makedirs(self.dataset_dir, exist_ok=True)
                    if self.writegjf:
                        os.makedirs(self.gjfdir, exist_ok=True)
                    self._writexyzfiles()
                gc.collect()
                timearray.append(time.time())
                logger.info(
                    f"Step {len(timearray)-1} Done! Time consumed (s): {timearray[-1]-timearray[-2]:.3f}"
                )

    def _readtimestepsbond(self):
        """Read and store the bond of each atom in each frame."""
        # added on 2018-12-15
        stepatomfiles = {}
        os.makedirs(self.trajatom_dir, exist_ok=True)
        results = run_mp(
            self.nproc,
            func=self.bonddetector.readatombondtype,
            l=zip(self.lineiter(self.bonddetector), self.erroriter())
            if self.errorfilename is not None
            else self.lineiter(self.bonddetector),
            return_num=True,
            extra=self.errorfilename is not None,
            desc="Read trajectory",
            unit="timestep",
        )
        nstep = 0
        for d, step in results:
            for bondtypebytes, atomids in d.items():
                bondtype = self._bondtype(bondtypebytes)
                if bondtype not in self.atombondtype:
                    self.atombondtype.append(bondtype)
                    stepatomfiles[bondtype] = open(
                        os.path.join(self.trajatom_dir, f"stepatom.{bondtype}"), "wb"
                    )
                stepatomfiles[bondtype].write(listtobytes([step, atomids]))
            nstep += 1
        self._nstep = nstep
        for stepatomfile in stepatomfiles.values():
            stepatomfile.close()

    def _writecoulumbmatrix(self, trajatomfilename, fc):
        """Write Coulumb matrix.

        Parameters
        ----------
        trajatomfilename : str
            The name of the bond, for example, C1111.
        fc : File object
            The File object for storing selected atoms.
        """
        self.dstep = {}
        with open(
            os.path.join(self.trajatom_dir, f"stepatom.{trajatomfilename}"), "rb"
        ) as f:
            for line in read_compressed_block(f):
                s = bytestolist(line)
                self.dstep[s[0]] = s[1]
        n_atoms = sum(map(len, self.dstep.values()))
        if n_atoms > self.n_clusters:
            # undersampling
            max_counter = Counter()
            stepatom = np.zeros((n_atoms, 2), dtype=int)
            feedvector = np.zeros((n_atoms, 0))
            vector_elements = defaultdict(list)
            results = run_mp(
                self.nproc,
                func=self._writestepmatrix,
                l=self.lineiter(self.crddetector),
                return_num=True,
                total=self._nstep,
                desc=trajatomfilename,
                unit="timestep",
            )
            j = 0
            for result in results:
                for stepatoma, vector, symbols_counter in result:
                    stepatom[j] = stepatoma
                    for element in (symbols_counter - max_counter).elements():
                        vector_elements[element].append(feedvector.shape[1])
                        feedvector = np.pad(
                            feedvector,
                            ((0, 0), (0, 1)),
                            "constant",
                            constant_values=(0, self._coulumbdiag[element]),
                        )
                    feedvector[
                        j,
                        sum(
                            (
                                vector_elements[x[0]][: x[1]]
                                for x in symbols_counter.items()
                            ),
                            [],
                        ),
                    ] = vector
                    max_counter |= symbols_counter
                    j += 1
            logger.info(f"Max counter of {trajatomfilename} is {max_counter}")
            choosedindexs = self._clusterdatas(
                np.sort(feedvector), n_clusters=self.n_clusters, n_each=self.n_each
            )
        else:
            stepatom = np.array([[u, vv] for u, v in self.dstep.items() for vv in v])
            choosedindexs = range(n_atoms)
        fc.write(listtobytes(stepatom[choosedindexs]))
        self._nstructure += len(choosedindexs)

    def _writestepmatrix(self, item):
        """Calculate Coulumb atoms for each atom.

        Parameters
        ----------
        item : tuple (step, lines)
            step: int
                The timestep of the frame.
            lines: list of strs
                Lines of the fram in the LAMMPS dump file.

        Returns
        -------
        results: list of tuples
            The tuple (stepatoma, columbmatrix, symbols) contains:
                stepatoma: numpy.ndarray (2,)
                    Contains two elements: step and atom ID.
                columbmatrix: numpy.ndarray (N,)
                    The eigenvalues of columb matrix.
                symbols: collections.Counter
                    The elements of atoms.
        """
        step, lines = item
        results = []
        if step in self.dstep:
            assert isinstance(self.crddetector, DetectDump)
            step_atoms, _ = self.crddetector.readcrd(lines)
            for atoma in self.dstep[step]:
                # atom ID starts from 1
                distances = step_atoms.get_distances(
                    atoma - 1, range(len(step_atoms)), mic=True
                )
                cutoffatoms = step_atoms[distances < self.cutoff]
                assert isinstance(cutoffatoms, Atoms)
                symbols = cutoffatoms.get_chemical_symbols()
                results.append(
                    (
                        np.array([step, atoma]),
                        self._calcoulumbmatrix(cutoffatoms),
                        Counter(symbols),
                    )
                )
        return results

    def _calcoulumbmatrix(self, atoms):
        """Calculate Coulumb matrix for atoms.

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms to calculate Coulumb matrix.

        Returns
        -------
        numpy.darray (N,)
            The eigenvalues of columb matrix.
        """
        # https://github.com/crcollins/molml/blob/master/molml/utils.py
        top = np.outer(atoms.numbers, atoms.numbers).astype(np.float64)
        r = atoms.get_all_distances(mic=True)
        diag = np.array(list(map(self._coulumbdiag.get, atoms.get_chemical_symbols())))
        with np.errstate(divide="ignore", invalid="ignore"):
            np.divide(top, r, top)
            np.fill_diagonal(top, diag)
        top[top == np.Infinity] = 0
        top[np.isnan(top)] = 0
        return np.linalg.eigh(top)[0]

    @classmethod
    def _clusterdatas(cls, X, n_clusters, n_each=1):
        """Select data using Mini Batch Kmeans.

        Parameters
        ----------
        X : numpy.darray
            The input data.
        n_clusters : int
            The number of clusters.
        n_each : int, optional, default=1
            The number of structures in each cluster.

        Returns
        -------
        numpy.ndarray
            The selected index.
        """
        min_max_scaler = preprocessing.MinMaxScaler()
        X = np.array(min_max_scaler.fit_transform(X))
        clus = MiniBatchKMeans(
            n_clusters=n_clusters,
            init_size=(min(3 * n_clusters, len(X))),
            n_init=3,  # type: ignore
        )
        labels = clus.fit_predict(X)
        choosedidx = []
        for i in range(n_clusters):
            idx = np.where(labels == i)[0]
            if idx.size:
                choosedidx.append(np.random.choice(idx, n_each))
        index = np.concatenate(choosedidx)
        return index

    def _writexyzfiles(self):
        """Write xyz files.

        Notes
        -----
        The functions writes a list to `self.dstep[step]`, where the list
        contains a tuple (atoma, trajatomfilename, itype, itotal):
            atoma: int
                the selected atom index (starts from 1)
            trajatomfilename: str
                the name of the bond, such as C1111
            itype: int
                index of structures in the bond type
            itotal: int
                index of structures in all
        """
        self.dstep = defaultdict(list)
        with open(os.path.join(self.trajatom_dir, "chooseatoms"), "rb") as fc:
            typecounter = Counter()
            for typefile, trajatomfilename in zip(
                read_compressed_block(fc), self.atombondtype
            ):
                for step, atoma in bytestolist(typefile):
                    self.dstep[step].append(
                        (
                            atoma,
                            trajatomfilename,
                            typecounter[trajatomfilename],
                            typecounter["total"],
                        )
                    )
                    typecounter[trajatomfilename] += 1
                    typecounter["total"] += 1
            self.maxlength = len(str(self.n_clusters))
            foldernum = self._nstructure // 1000 + 1
            self.foldermaxlength = len(str(foldernum))
            foldernames = [str(i).zfill(self.foldermaxlength) for i in range(foldernum)]
            for folder in foldernames:
                os.makedirs(os.path.join(self.dataset_dir, folder), exist_ok=True)
            if self.writegjf:
                for folder in foldernames:
                    os.makedirs(os.path.join(self.gjfdir, folder), exist_ok=True)
            crditer = self.lineiter(self.crddetector)
            if self.crddetector is self.bonddetector:
                lineiter = crditer
            else:
                bonditer = self.lineiter(self.bonddetector)
                lineiter = zip(crditer, bonditer)
            results = run_mp(
                self.nproc,
                func=self._writestepxyzfile,
                l=lineiter,
                return_num=True,
                total=self._nstructure,
                desc="Write structures",
                unit="structure",
            )
            for _ in results:
                pass

    @staticmethod
    def detect_multiplicity(symbols):
        """Caculate multiplicity.

        Parameters
        ----------
        symbols : numpy.ndarray
            The atomic symbols.

        Returns
        -------
        multiplicity: int
            The multiplicity.
        """
        # currently only support charge=0
        # oxygen -> 3
        if symbols == ["O", "O"]:
            return 3
        # calculates the total number of electrons, assumes they are paired as much as possible
        n_total = sum([atomic_numbers[s] for s in symbols])
        return n_total % 2 + 1

    def _convertgjf(self, gjffilename, takenatomidindex, atoms_whole):
        """Generate GJF files.

        Parameters
        ----------
        gjffilename : str
            The filename of GJF file.
        takenatomidindex : list
            The index of taken atoms.
        atoms_whole : ase.Atoms
            The whole atoms in the frame.
        """
        buff = []
        multiplicities = [
            self.detect_multiplicity(atoms_whole[atoms].get_chemical_symbols())
            for atoms in takenatomidindex
        ]
        multiplicity_whole = sum(multiplicities) - len(takenatomidindex) + 1
        multiplicity_whole_str = f"0 {multiplicity_whole}"
        title = "\nGenerated by MDDatasetMaker (Author: Jinzhe Zeng)\n"
        connect = None
        if len(self.qmkeywords) > 1:
            connect = "\n--link1--\n"
            chk = [f"%chk={os.path.splitext(os.path.basename(gjffilename))[0]}.chk"]
        else:
            chk = []
        if len(takenatomidindex) == 1 or not self.fragment:
            buff.extend((*chk, self.qmkeywords[0], title, multiplicity_whole_str))
            buff.extend(
                "{} {:.5f} {:.5f} {:.5f}".format(atom.symbol, *atom.position)
                for atom in atoms_whole
            )
        else:
            kw0 = f"{self.qmkeywords[0]} guess=fragment={len(takenatomidindex)}"
            multiplicities_str = "{} {}".format(
                multiplicity_whole_str,
                " ".join([f"0 {multiplicity}" for multiplicity in multiplicities]),
            )
            buff.extend((*chk, kw0, title, multiplicities_str))
            for index, atoms in enumerate(takenatomidindex, 1):
                buff.extend(
                    [
                        "{}(Fragment={}) {:.5f} {:.5f} {:.5f}".format(
                            atom.symbol, index, *atom.position
                        )
                        for atom in atoms_whole[atoms]
                    ]
                )
        for kw in itertools.islice(self.qmkeywords, 1, None):
            assert connect is not None
            buff.extend((connect, *chk, kw, title, f"0 {multiplicity_whole}", "\n"))
        buff.append("\n")
        with open(gjffilename, "w") as f:
            f.write("\n".join(buff))

    def _writestepxyzfile(self, item):
        """Write xyz files and GJF files in a timestep.

        Parameters
        ----------
        item : tuple (step, lines)
            step: int
                The timestep of the frame.
            lines: list of strs or list of lists of strs
                Lines of the fram in the LAMMPS dump file (and bond file).

        Returns
        -------
        results: int
            The number of written files.
        """
        step, lines = item
        results = 0
        if step in self.dstep:
            if len(lines) == 2:
                assert isinstance(self.crddetector, DetectDump)
                step_atoms, _ = self.crddetector.readcrd(lines[0])
                molecules, _ = self.bonddetector.readmolecule(lines[1])
            else:
                molecules, step_atoms = self.bonddetector.readmolecule(lines)
            assert step_atoms is not None
            for atoma, trajatomfilename, itype, itotal in self.dstep[step]:
                # update counter
                folder = str(itotal // 1000).zfill(self.foldermaxlength)
                atomtypenum = str(itype).zfill(self.maxlength)
                # atom ID starts from 1
                distances = step_atoms.get_distances(
                    atoma - 1, range(len(step_atoms)), mic=True
                )
                cutoffatomid = np.where(distances < self.cutoff)
                # make cutoff atoms in molecules
                takenatomids = []
                takenatomidindex = []
                idsum = 0
                for mo in molecules:
                    mol_atomid = np.array(mo)
                    if np.any(np.isin(mol_atomid, cutoffatomid)):
                        takenatomids.append(mol_atomid)
                        takenatomidindex.append(range(idsum, idsum + len(mol_atomid)))
                        idsum += len(mol_atomid)
                idx = np.concatenate(takenatomids)
                cutoffatoms = step_atoms[idx]
                assert isinstance(cutoffatoms, Atoms)
                cutoffatoms[np.nonzero(idx == atoma - 1)[0][0]].tag = 1  # type: ignore
                cutoffatoms.wrap(
                    center=step_atoms[atoma - 1].position  # type: ignore
                    / cutoffatoms.get_cell_lengths_and_angles()[0:3],
                    pbc=cutoffatoms.get_pbc(),
                )
                write_xyz(
                    os.path.join(
                        self.dataset_dir,
                        folder,
                        f"{self.xyzfilename}_{trajatomfilename}_{atomtypenum}.xyz",
                    ),
                    cutoffatoms,
                    format="xyz",
                )
                if self.writegjf:
                    self._convertgjf(
                        os.path.join(
                            self.gjfdir,
                            folder,
                            f"{self.xyzfilename}_{trajatomfilename}_{atomtypenum}.gjf",
                        ),
                        takenatomidindex,
                        cutoffatoms,
                    )
                if self.atom_pref:
                    np.save(
                        os.path.join(
                            self.gjfdir,
                            folder,
                            f"{self.xyzfilename}_{trajatomfilename}_{atomtypenum}.atom_pref.npy",
                        ),
                        np.array([cutoffatoms.get_tags()]),
                    )
                results += 1
        return results

    def _bondtype(self, typebytes):
        if typebytes in self.bondtyperestore:
            return self.bondtyperestore[typebytes]
        typetuple = pickle.loads(typebytes)
        typestr = f"{typetuple[0]}{''.join(map(str,typetuple[1]))}"
        self.bondtyperestore[typebytes] = typestr
        return typestr

    def lineiter(self, detector):
        """Iterate over file(s).

        Parameters
        ----------
        detector : mddatasetbuilder.detect.Detect
            File detector

        Yields
        ------
        str
            lines
        """
        fns = must_be_list(detector.filename)
        for fn in fns:
            with open(fn) as f:
                it = itertools.islice(
                    itertools.zip_longest(*[f] * detector.steplinenum),
                    0,
                    None,
                    self.stepinterval,
                )
                yield from it

    def erroriter(self):
        """Iterate over the model deviation file.

        Yields
        ------
        str
            the line of model deviation
        """
        assert self.errorfilename is not None
        fns = must_be_list(self.errorfilename)
        for fn in fns:
            with open(fn) as f:
                it = itertools.islice(f, 1, None)
                yield from it


def _commandline():
    parser = argparse.ArgumentParser(
        description="MDDatasetBuilder",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--dumpfile",
        nargs="*",
        help="Input dump file, e.g. dump.reaxc",
        required=True,
    )
    parser.add_argument(
        "-b",
        "--bondfile",
        nargs="*",
        help="Input bond file, e.g. bonds.reaxc. If not given, Open Babel will be used to determine bond connectivity.",
    )
    parser.add_argument(
        "-a",
        "--atomname",
        help="Atomic names in the trajectory, e.g. C H O",
        nargs="*",
        required=True,
    )
    parser.add_argument(
        "-np", "--nproc", help="Number of processes used by MDDatasetBuilder", type=int
    )
    parser.add_argument(
        "-c",
        "--cutoff",
        help="Cutoff radius to take clusters from the trajectory. Note that only a complete molecule or free radical will be taken.",
        type=float,
        default=5.0,
    )
    parser.add_argument(
        "-i",
        "--interval",
        help="Step interval to collect from the trajectory. 1 means collect from every step in the trajectory.",
        type=int,
        default=1,
    )
    parser.add_argument(
        "-s",
        "--size",
        help="Collected dataset size for each bond type.",
        type=int,
        default=10000,
    )
    parser.add_argument(
        "-k",
        "--qmkeywords",
        help='Gaussian QM keywords. Note that it should include "force" keyword to compute forces. Add "Geom=PrintInputOrient" if the number of atoms will be more than 50.',
        default="force mn15/6-31g** Geom=PrintInputOrient",
    )
    parser.add_argument(
        "--nprocjob",
        help="CPU number that each Gaussian job uses.",
        default=4,
        type=int,
    )
    parser.add_argument("-n", "--name", help="Dataset name", default="md")
    parser.add_argument(
        "--errorfile",
        help="Model deviation file generated by DeePMD-kit plugin for LAMMPS. atomic should be enabled.",
        nargs="*",
    )
    parser.add_argument(
        "-e",
        "--errorlimit",
        help="Model deviation lower threshold. Atoms whose model deviation is less than the threshold will not be collected.",
        type=float,
        default=0.0,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"MDDatasetBuilder {__version__}",
    )
    args = parser.parse_args()
    DatasetBuilder(
        atomname=args.atomname,
        bondfilename=args.bondfile,
        dumpfilename=args.dumpfile,
        dataset_name=args.name,
        cutoff=args.cutoff,
        stepinterval=args.interval,
        n_clusters=args.size,
        qmkeywords=f"%nproc={args.nprocjob}\n#{args.qmkeywords}",
        nproc=args.nproc,
        errorfilename=args.errorfile,
        errorlimit=args.errorlimit,
    ).builddataset()



"""Test command line interface."""

import subprocess as sp
import sys


def test_module():
    """Test python -m mddatasetbuilder."""
    sp.check_output([sys.executable, "-m", "mddatasetbuilder", "-h"])



import logging

import coloredlogs

from ._version import version as __version__

logger = logging.getLogger(__name__)

coloredlogs.install(
    fmt=f"%(asctime)s - MDDatasetBuilder {__version__} - %(levelname)s: %(message)s",
    level=logging.INFO,
    milliseconds=True,
    logger=logger,
)



# distutils: language = c++
# cython: language_level=3
# cython: linetrace=True
# cython: infer_types=True
"""Connect molecule with Depth-First Search."""
from libc.stdlib cimport malloc, free

import cython

cdef extern from "c_stack.h":
    # This function is copied from https://zhuanlan.zhihu.com/p/38212302
    cdef cppclass C_Stack:
        void push(int val)
        int pop()


@cython.binding(False)
def dps(bonds):
    molecule = []
    cdef int _N = len(bonds)
    cdef int *visited = <int *> malloc(_N * sizeof(int))
    cdef int i, s, b_c, nbonds, ib
    cdef C_Stack st
    for i in range(_N):
        visited[i]=0

    for i in range(_N):
        if visited[i]==0:
            mol = []
            st.push(i)
            while True:
                s = st.pop()
                if s < 0:
                    break
                elif visited[s]==1:
                    continue
                mol.append(s)
                nbonds = len(bonds[s])
                for ib in range(nbonds):
                    b_c = bonds[s][ib]
                    if visited[b_c]==0:
                        st.push(b_c)
                visited[s]=1
            molecule.append(mol)
    free(visited)
    return molecule
